# -*- coding: utf-8 -*-

# ╔════════════════════════════════════════════════════════════════════╗
# ║   LINIAROTE: The programming language for performing operations    ║
# ║   in transvalent mathematics                                       ║
# ║                                                                    ║
# ║   Developed by Matthew E. Gladden                                  ║
# ║   Software and documentation ©2022-2023 Cognitive Firewall LLC     ║
# ║                                                                    ║
# ║   This software is made available for use under                    ║
# ║   GNU General Public License Version 3                             ║
# ║   (please see https://www.gnu.org/licenses/gpl-3.0.html).          ║
# ╚════════════════════════════════════════════════════════════════════╝

"""
This module stores configuration settings and variables that are used by
other modules.
"""

# ██████████████████████████████████████████████████████████████████████
# ██████████████████████████████████████████████████████████████████████
# ███
# ███ DEFINE VARIABLES
# ███
# ██████████████████████████████████████████████████████████████████████
# ██████████████████████████████████████████████████████████████████████

# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Transvalent symbols
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

transval_sym_pos = "Ƿ"
transval_sym_neg = "-" + transval_sym_pos


# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Text for display
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

# ----------------------------------------------------------------------
# Introductory text to be displayed whenever the package is run.
# ----------------------------------------------------------------------

intro_text = \
"""
-----------------------------------------------------------------------------
****** LINIAROTE: The programming language for transvalent mathematics ******
To enter transvalent numbers, use the letter 'w' to represent the 'Ƿ' symbol.
For more detailed instructions, please type 'help’ at the command prompt.
-----------------------------------------------------------------------------
"""


# ----------------------------------------------------------------------
# Help text to be displayed whenever requested by the user.
# ----------------------------------------------------------------------

help_text = \
"""
-----------------------------------------------------------------------------
The Liniarote package is a Python implementation of a command-line interface
for the Liniarote programming language, an open-source language developed 
to facilitate operations in transvalent mathematics.

>>>>> AN OVERVIEW OF TRANSVALENT MATHEMATICS <<<<<

The distinguishing characteristic of transvalent mathematics is that division 
by zero is permitted and generates results calculated according to certain 
clearly defined principles.

The symbol “Ƿ” is used to represent that (transvalent) number which when 
multiplied by 0 yields a result of 1. An equivalent way of defining Ƿ is to 
state that it is the quotient generated when 1 is divided by 0.

Informally, one can think of the number Ƿ as a special type of “infinity” 
that possesses such transcendent power that when 0 is multiplied by it, Ƿ is 
capable of “lifting” zero out of nothingness to give their product a positive 
value (i.e., 1). Similarly, the symbol “-Ƿ” is used to represent that 
(transvalent) number which when multiplied by 0 yields a result of -1; 
alternatively, -Ƿ can be defined as the quotient generated when -1 is divided 
by 0.

The number Ƿ is described as “imperishable,” because there is no real number 
that one can multiply it by in order to yield 0. (Similarly, there is no real 
number by which -Ƿ can be multiplied to yield 0.)

Intuitively, it is easy to suppose that attempts at dividing by zero must 
necessarily yield a result that is “undefined,” “meaningless,” or “absurd” 
and that calculations that attempt to divide by zero are inherently 
“erroneous.” From a theoretical perspective, though, the definitions that 
“Ƿ = 1 ÷ 0” and “Ƿ × 0 = 1” are no more intrinsically absurd than (for 
example) the definitions that “i = √(-1)” and “i² = -1”, which are 
foundations of the imaginary and complex number systems – and which have 
facilitated countless theoretical and practical advances in diverse 
scientific and technological spheres. The establishment of systems like that 
found in transvalent mathematics simply required clear formulations of the 
meaning of “dividing by zero” and rules for interpreting the results that are
generated. Frameworks with similarities to transvalent mathematics (e.g., 
in granting some formal meaning to division by zero) that have been developed 
in the last 200 years include those of the Riemann sphere (which extends the 
complex plane by adding a point at infinity) and the projectively extended 
real line.

>>>>> SELECTED AXIOMS OF TRANSVALENT MATHEMATICS <<<<<

   Axioms of identity:

   --->    Ƿ = 0 + Ƿ
   --->   -Ƿ = 0 - Ƿ
   --->    n = n + (Ƿ - Ƿ), where n is any real number

   Axioms of non-associativity:

   --->    (Ƿ + Ƿ) - Ƿ = 0
   --->    (Ƿ + (Ƿ - Ƿ) = Ƿ

   Axioms of addition and subtraction:

   --->    Ƿ + Ƿ = Ƿ
   --->   -Ƿ - Ƿ = -Ƿ
   --->    Ƿ - Ƿ = 0
   --->   -Ƿ + Ƿ = 0

   Axioms of multiplication:

   --->    Ƿ × 0 = 1
   --->   -Ƿ × 0 = -1

   --->    Ƿ × n = Ƿ, where n is any positive real number
   --->    Ƿ × n = -Ƿ, where n is any negative real number
   --->   -Ƿ × n = -Ƿ, where n is any positive real number
   --->   -Ƿ × n = Ƿ, where n is any negative real number
   --->    Ƿ × n ≠ 0, where n is any real number
   --->   -Ƿ × n ≠ 0, where n is any real number

   --->    Ƿ × Ƿ = Ƿ
   --->   -Ƿ × Ƿ = -Ƿ
   --->   -Ƿ × -Ƿ = Ƿ

   Axioms of division:

   --->    n ÷ 0 = Ƿ, where n is any positive real number
   --->    n ÷ 0 = -Ƿ, where n is any negative real number
   --->    0 ÷ 0 = 1
   --->    Ƿ ÷ 0 = Ƿ
   --->   -Ƿ ÷ 0 = -Ƿ

   --->    Ƿ ÷ n = Ƿ, where n is any positive real number
   --->    Ƿ ÷ n = -Ƿ, where n is any negative real number
   --->   -Ƿ ÷ n = -Ƿ, where n is any positive real number
   --->   -Ƿ ÷ n = Ƿ, where n is any negative real number

   --->    n ÷ Ƿ = 0, where n is any positive real number
   --->    n ÷ Ƿ = -1, where n is any negative real number
   --->    n ÷ -Ƿ = -1, where n is any positive real number
   --->    n ÷ -Ƿ = 0, where n is any negative real number

   --->    0 ÷ Ƿ = 0
   --->    0 ÷ -Ƿ = -1

   --->    Ƿ ÷ Ƿ = Ƿ
   --->   -Ƿ ÷ Ƿ = -Ƿ
   --->   -Ƿ ÷ -Ƿ = Ƿ
   --->    Ƿ ÷ -Ƿ = -Ƿ

>>>>> PACKAGE EXECUTION AND INPUT FORMAT <<<<<

The manner of activating the Liniarote command-line interface will depend on
one's operating system and Python installation. In Windows, for example, one
may be able to run the command-line interface module (cli.py) by typing at
the PowerShell prompt or Visual Studio Code terminal prompt from within the
appropriate folder:

   python -m pip install liniarote
   python -m liniarote.cli

On the Liniarote command line, the symbol “Ƿ” (Unicode: U+01F7; UTF-8: C7 B7) 
can be inputted by typing the lowercase letter “w”, which will be 
automatically converted into “Ƿ” during processing of the input. (Similarly, 
“-w” will be converted into “-Ƿ”.)

The order of entry of the real and transvalent components of a number is 
without significance. Thus, both “3 + w” and “w + 3” will, when inputted, be 
converted into “3 + Ƿ”, while “-5 + w” and “-w – 5” and will be converted 
into “-5 – Ƿ”.

Inputted constants recognized by Liniarote include “e” and “pi”.

Division is not yet implemented. The operators currently available for use 
are: + - * ( )

>>>>> REQUIREMENTS <<<<<

Please see the “requirements.txt” file for a list of other Python packages 
whose installation is a prerequisite for the proper functioning of Liniarote.

>>>>> INSPIRATION AND ACKNOWLEDGMENTS <<<<<

Within the world of the “Utopian Confederation” RPG series, the foundational 
text in the field of transvalent mathematics as such was published by a team 
of Utopian mathematicians in 1911. In honor of their work, the Liniarote 
programming language was named after one of the 54 cities of the ancient 
Utopian Commonwealth, as described in ThomasMore’s “Utopia” (1516). Liniarote 
was created and is maintained by Matthew E. Gladden.

Liniarote code and documentation ©2022-2023 Cognitive Firewall LLC
-----------------------------------------------------------------------------
"""

# ••••-••••-••••-••••-••••-••••-••••--••••-••••-••••-••••-••••-••••-••••


# -*- coding: utf-8 -*-

# ╔════════════════════════════════════════════════════════════════════╗
# ║   LINIAROTE: The programming language for performing operations    ║
# ║   in transvalent mathematics                                       ║
# ║                                                                    ║
# ║   Developed by Matthew E. Gladden                                  ║
# ║   Software and documentation ©2022-2023 Cognitive Firewall LLC     ║
# ║                                                                    ║
# ║   This software is made available for use under                    ║
# ║   GNU General Public License Version 3                             ║
# ║   (please see https://www.gnu.org/licenses/gpl-3.0.html).          ║
# ╚════════════════════════════════════════════════════════════════════╝

"""
This module handles all of the programming language's input/output
mechanisms and internal logic.
"""

# ██████████████████████████████████████████████████████████████████████
# ██████████████████████████████████████████████████████████████████████
# ███
# ███ PRELIMINARY STEPS
# ███
# ██████████████████████████████████████████████████████████████████████
# ██████████████████████████████████████████████████████████████████████

# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Import standard modules
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

import time
import math
import signal


# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Import third-party modules
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

import readchar
from sly import Lexer, Parser


# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Import other modules from the WorkforceSim package
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

try:
    from . import config as cfg
except:
    import config as cfg


# ██████████████████████████████████████████████████████████████████████
# ██████████████████████████████████████████████████████████████████████
# ███
# ███ DEFINE CLASSES AND FUNCTIONS
# ███
# ██████████████████████████████████████████████████████████████████████
# ██████████████████████████████████████████████████████████████████████

# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Lexer and Parser
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

class LiniaroteLexer(Lexer):
    """
    Liniarote's core lexer, for recognizing inputted characters as tokens.
    """

    # Tokens to be recognized.
    tokens = {
        PLUS,
        MINUS,
        TIMES,
        #DIVIDE,
        LPAREN,
        RPAREN,
        NUM,
        ID,
        TRANSVALENT_SYMBOL_INPUT,
        HELP,
        }

    # Input to be ignored.
    ignore = ' \t\r\n'

    # Character strings to be recognized as tokens.
    PLUS = r'\+'
    MINUS = r'-'
    TIMES = r'\*'
    #DIVIDE = r'/'
    LPAREN = r'\('
    RPAREN = r'\)'
    NUM = r"-?[0-9]+(\.[0-9]*)?"
    ID = r"[a-zA-Z_][a-zA-Z0-9_]*"
    ID['w'] = TRANSVALENT_SYMBOL_INPUT
    ID['W'] = TRANSVALENT_SYMBOL_INPUT
    ID['help'] = HELP
    ID['?'] = HELP


class LiniaroteParser(Parser):
    """
    Liniarote's core parser, for performing operations on tokenized input.
    """

    # Uncomment the line below to generate a debug file (e.g., for 
    # debugging shift/reduce or reduce/reduce conflicts).
    #debugfile = "parser.out"

    # ------------------------------------------------------------------
    # Specify tokens and the order of operations.
    # ------------------------------------------------------------------

    tokens = LiniaroteLexer.tokens

    # Specify the order in which operations should be processed.
    precedence = (
        ('left', PLUS, MINUS),
        #('left', TIMES, DIVIDE),
        ('left', TIMES),
        ('right', UMINUS),
        )


    # ------------------------------------------------------------------
    # Define internal functions.
    # ------------------------------------------------------------------

    def __init__(self):
        """
        Constructor for the class object.
        """
        self.vars = {
            "pi": math.pi,
            "e": math.e,
            }


    @_('HELP')
    def expr(self, p):
        """
        Displays help text when the HELP token is recognized.
        """
        return print_help_text()


    @_('MINUS expr %prec UMINUS')
    def expr(self, p):
        """
        Gives precedence to the processing of a unary minus sign
        (over the subtraction operation).
        """

        if p.expr == cfg.transval_sym_pos:
            return cfg.transval_sym_neg
        if p.expr == cfg.transval_sym_neg:
            return cfg.transval_sym_pos
        return -p.expr


    @_('expr MINUS expr')
    def expr(self, p):
        """
        Specifies how the subtraction operation is evaluated
        within the context of transvalent mathematics.
        """

        # ------------------------------------------------------------------
        # Process subtraction between lone elements (e.g., a transvalent 
        # symbol *or* float with another such variable), generating a
        # well-formed transvalent tuple as output.
        # ------------------------------------------------------------------

        # Process "float - float".
        if isinstance(p.expr0, float) and isinstance(p.expr1, float):
            return (p.expr0 - p.expr0, 0.0)

        # Process "float - positive transvalent".
        elif isinstance(p.expr0, float) and (p.expr1 == cfg.transval_sym_pos):
            return (p.expr0, cfg.transval_sym_neg)

        # Process "float - negative transvalent".
        elif isinstance(p.expr0, float) and (p.expr1 == cfg.transval_sym_neg):
            return (p.expr0, cfg.transval_sym_pos)

        # Process "positive transvalent - float".
        elif (p.expr0 == cfg.transval_sym_pos) and isinstance(p.expr1, float):
            return (-p.expr1, cfg.transval_sym_pos)

        # Process "negative transvalent - float".
        elif (p.expr0 == cfg.transval_sym_neg) and isinstance(p.expr1, float):
            return (-p.expr1, cfg.transval_sym_neg)

        # Process "positive transvalent - positive transvalent".
        elif (p.expr0 == cfg.transval_sym_pos) and (p.expr1 == cfg.transval_sym_pos):
            return (0.0, 0.0)

        # Process "positive transvalent - negative transvalent".
        elif (p.expr0 == cfg.transval_sym_pos) and (p.expr1 == cfg.transval_sym_neg):
            return (0.0, cfg.transval_sym_pos)

        # Process "negative transvalent - positive transvalent".
        elif (p.expr0 == cfg.transval_sym_neg) and (p.expr1 == cfg.transval_sym_pos):
            return (0.0, cfg.transval_sym_neg)

        # Process "negative transvalent - negative transvalent".
        elif (p.expr0 == cfg.transval_sym_neg) and (p.expr1 == cfg.transval_sym_neg):
            return (0.0, 0.0)


        # ------------------------------------------------------------------
        # Process subtraction between a lone element and a tuple, generating
        # a well-formed transvalent tuple as output.
        # ------------------------------------------------------------------

        # Process "float - tuple".
        elif isinstance(p.expr0, float) and isinstance(p.expr1, tuple):
            if p.expr1[1] == cfg.transval_sym_pos:
                transvalent_elements_diff = cfg.transval_sym_neg
            elif p.expr1[1] == cfg.transval_sym_neg:
                transvalent_elements_diff = cfg.transval_sym_pos
            elif p.expr1[1] == 0:
                transvalent_elements_diff = 0.0
            return (p.expr0 - p.expr1[0], transvalent_elements_diff)

        # Process "tuple - float".
        elif isinstance(p.expr0, tuple) and isinstance(p.expr1, float):
            if p.expr0[1] == cfg.transval_sym_pos:
                transvalent_elements_diff = cfg.transval_sym_pos
            elif p.expr0[1] == cfg.transval_sym_neg:
                transvalent_elements_diff = cfg.transval_sym_neg
            elif p.expr0[1] == 0:
                transvalent_elements_diff = 0.0
            return (p.expr0[0] - p.expr1, transvalent_elements_diff)

        # Process "transvalent - tuple".
        elif (p.expr0 == cfg.transval_sym_pos) and isinstance(p.expr1, tuple):
            if p.expr1[1] == cfg.transval_sym_pos:
                transvalent_elements_diff = 0.0
            elif p.expr1[1] == cfg.transval_sym_neg:
                transvalent_elements_diff = cfg.transval_sym_pos
            elif p.expr1[1] == 0:
                transvalent_elements_diff = cfg.transval_sym_pos
            return (p.expr0[0] - p.expr1, transvalent_elements_diff)

        elif (p.expr0 == cfg.transval_sym_neg) and isinstance(p.expr1, tuple):
            if p.expr1[1] == cfg.transval_sym_pos:
                transvalent_elements_diff = cfg.transval_sym_neg
            elif p.expr1[1] == cfg.transval_sym_neg:
                transvalent_elements_diff = 0.0
            elif p.expr1[1] == 0:
                transvalent_elements_diff = cfg.transval_sym_neg
            return (p.expr0[0] - p.expr1, transvalent_elements_diff)

        # NOTE: this is unnecessary, as it is already handled by
        # the case of "float - tuple".
        #
        #elif (p.expr0 == 0) and isinstance(p.expr1, tuple):
        #    if p.expr1[1] == cfg.transval_sym_pos:
        #        transvalent_elements_diff = cfg.transval_sym_neg
        #    elif p.expr1[1] == cfg.transval_sym_neg:
        #        transvalent_elements_diff = cfg.transval_sym_pos
        #    elif p.expr1[1] == 0:
        #        transvalent_elements_diff = 0.0
        #    return (p.expr0[0] - p.expr1, transvalent_elements_diff)

        # Process "tuple - transvalent".
        elif isinstance(p.expr0, tuple) and (p.expr1 == cfg.transval_sym_pos):
            if p.expr0[1] == cfg.transval_sym_pos:
                transvalent_elements_diff = 0.0
            elif p.expr0[1] == cfg.transval_sym_neg:
                transvalent_elements_diff = cfg.transval_sym_neg
            elif p.expr0[1] == 0:
                transvalent_elements_diff = cfg.transval_sym_neg
            return (p.expr0[0], transvalent_elements_diff)

        elif isinstance(p.expr0, tuple) and (p.expr1 == cfg.transval_sym_neg):
            if p.expr0[1] == cfg.transval_sym_pos:
                transvalent_elements_diff = cfg.transval_sym_pos
            elif p.expr0[1] == cfg.transval_sym_neg:
                transvalent_elements_diff = 0.0
            elif p.expr0[1] == 0:
                transvalent_elements_diff = cfg.transval_sym_pos
            return (p.expr0[0], transvalent_elements_diff)

        # NOTE: this is unnecessary, as it is already handled by
        # the case of "tuple - float".
        #
        #elif isinstance(p.expr0, tuple) and (p.expr1 == 0):
        #    if p.expr0[1] == cfg.transval_sym_pos:
        #        transvalent_elements_diff = cfg.transval_sym_pos
        #    elif p.expr0[1] == cfg.transval_sym_neg:
        #        transvalent_elements_diff = cfg.transval_sym_neg
        #    elif p.expr0[1] == 0:
        #        transvalent_elements_diff = 0.0
        #    return (p.expr0[0], transvalent_elements_diff)


        # ------------------------------------------------------------------
        # Process subtraction between two tuples.
        # ------------------------------------------------------------------

        elif isinstance(p.expr0, tuple) and isinstance(p.expr1, tuple):

            # Determine what the difference of the transvalent elements will be.
            if (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == cfg.transval_sym_pos):
                transvalent_elements_diff = 0.0
            elif (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == cfg.transval_sym_neg):
                transvalent_elements_diff = cfg.transval_sym_pos
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == cfg.transval_sym_neg):
                transvalent_elements_diff = 0.0
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == cfg.transval_sym_pos):
                transvalent_elements_diff = cfg.transval_sym_neg

            elif (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == 0):
                transvalent_elements_diff = cfg.transval_sym_pos
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == 0):
                transvalent_elements_diff = cfg.transval_sym_neg
            elif (p.expr0[1] == 0) and (p.expr0[1] == 0):
                transvalent_elements_diff = 0
            elif (p.expr0[1] == 0) and (p.expr0[1] == cfg.transval_sym_pos):
                transvalent_elements_diff = cfg.transval_sym_neg
            elif (p.expr0[1] == 0) and (p.expr0[1] == cfg.transval_sym_neg):
                transvalent_elements_diff = cfg.transval_sym_pos

            return (p.expr0[0] - p.expr1[0], transvalent_elements_diff)


    @_('expr PLUS expr')
    def expr(self, p):
        """
        Specifies how the addition operation is evaluated
        within the context of transvalent mathematics.
        """

        # ------------------------------------------------------------------
        # Process addition of lone elements (e.g., a transvalent 
        # symbol *or* float with another such variable), generating a
        # well-formed transvalent tuple as output.
        # ------------------------------------------------------------------

        # Process "float + float".
        if isinstance(p.expr0, float) and isinstance(p.expr1, float):
            return (p.expr0 + p.expr0, 0.0)

        # Process "float + positive transvalent".
        elif isinstance(p.expr0, float) and (p.expr1 == cfg.transval_sym_pos):
            return (p.expr0, cfg.transval_sym_pos)

        # Process "float + negative transvalent".
        elif isinstance(p.expr0, float) and (p.expr1 == cfg.transval_sym_neg):
            return (p.expr0, cfg.transval_sym_neg)

        # Process "positive transvalent + float".
        elif (p.expr0 == cfg.transval_sym_pos) and isinstance(p.expr1, float):
            return (p.expr1, cfg.transval_sym_pos)

        # Process "negative transvalent + float".
        elif (p.expr0 == cfg.transval_sym_neg) and isinstance(p.expr1, float):
            return (p.expr1, cfg.transval_sym_neg)

        # Process "positive transvalent + positive transvalent".
        elif (p.expr0 == cfg.transval_sym_pos) and (p.expr1 == cfg.transval_sym_pos):
            return (0.0, cfg.transval_sym_pos)

        # Process "positive transvalent + negative transvalent".
        elif (p.expr0 == cfg.transval_sym_pos) and (p.expr1 == cfg.transval_sym_neg):
            return (0.0, 0.0)

        # Process "negative transvalent + positive transvalent".
        elif (p.expr0 == cfg.transval_sym_neg) and (p.expr1 == cfg.transval_sym_pos):
            return (0.0, 0.0)

        # Process "negative transvalent + negative transvalent".
        elif (p.expr0 == cfg.transval_sym_neg) and (p.expr1 == cfg.transval_sym_neg):
            return (0.0, cfg.transval_sym_neg)


        # ------------------------------------------------------------------
        # Process addition of a lone element and a tuple, generating
        # a well-formed transvalent tuple as output.
        # ------------------------------------------------------------------

        # Process "float + tuple".
        elif isinstance(p.expr0, float) and isinstance(p.expr1, tuple):
            return (p.expr0 + p.expr1[0], p.expr1[1])

        # Process "tuple + float".
        elif isinstance(p.expr0, tuple) and isinstance(p.expr1, float):
            return (p.expr0[0] + p.expr1, p.expr0[1])

        # Process "transvalent + tuple".
        elif (p.expr0 == cfg.transval_sym_pos) and isinstance(p.expr1, tuple):
            if p.expr1[1] == cfg.transval_sym_pos:
                transvalent_elements_sum = cfg.transval_sym_pos
            if p.expr1[1] == cfg.transval_sym_neg:
                transvalent_elements_sum = 0.0
            if p.expr1[1] == 0:
                transvalent_elements_sum = cfg.transval_sym_pos
            return (p.expr1[0], transvalent_elements_sum)

        elif (p.expr0 == cfg.transval_sym_neg) and isinstance(p.expr1, tuple):
            if p.expr1[1] == cfg.transval_sym_pos:
                transvalent_elements_sum = 0.0
            if p.expr1[1] == cfg.transval_sym_neg:
                transvalent_elements_sum = cfg.transval_sym_neg
            if p.expr1[1] == 0:
                transvalent_elements_sum = cfg.transval_sym_neg
            return (p.expr1[0], transvalent_elements_sum)

        # NOTE: this is unnecessary, as it is already handled by
        # the case of "float + tuple".
        #
        #elif (p.expr0 == 0) and isinstance(p.expr1, tuple):
        #    if p.expr1[1] == cfg.transval_sym_pos:
        #        transvalent_elements_sum = cfg.transval_sym_pos
        #    if p.expr1[1] == cfg.transval_sym_neg:
        #        transvalent_elements_sum = cfg.transval_sym_neg
        #    if p.expr1[1] == 0:
        #        transvalent_elements_sum = 0.0
        #    return (p.expr1[0], transvalent_elements_sum)

        # Process "tuple + transvalent".
        elif isinstance(p.expr0, tuple) and (p.expr1 == cfg.transval_sym_pos):
            if p.expr0[1] == cfg.transval_sym_pos:
                transvalent_elements_sum = cfg.transval_sym_pos
            elif p.expr0[1] == cfg.transval_sym_neg:
                transvalent_elements_sum = 0.0
            elif p.expr0[1] == 0:
                transvalent_elements_sum = cfg.transval_sym_pos
            return (p.expr0[0], transvalent_elements_sum)

        elif isinstance(p.expr0, tuple) and (p.expr1 == cfg.transval_sym_neg):
            if p.expr0[1] == cfg.transval_sym_pos:
                transvalent_elements_sum = 0.0
            elif p.expr0[1] == cfg.transval_sym_neg:
                transvalent_elements_sum = cfg.transval_sym_neg
            elif p.expr0[1] == 0:
                transvalent_elements_sum = cfg.transval_sym_neg
            return (p.expr0[0], transvalent_elements_sum)

        # NOTE: this is unnecessary, as it is already handled by
        # the case of "tuple + float".
        #
        #elif isinstance(p.expr0, tuple) and (p.expr1 == 0):
        #    if p.expr0[1] == cfg.transval_sym_pos:
        #        transvalent_elements_sum = cfg.transval_sym_pos
        #    if p.expr0[1] == cfg.transval_sym_neg:
        #        transvalent_elements_sum = cfg.transval_sym_neg
        #    if p.expr0[1] == 0:
        #        transvalent_elements_sum = 0.0
        #    return (p.expr0[0], transvalent_elements_sum)


        # ------------------------------------------------------------------
        # Process addition of two tuples.
        # ------------------------------------------------------------------

        elif isinstance(p.expr0, tuple) and isinstance(p.expr1, tuple):

            # Determine what the sum of the transvalent elements will be.
            if (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == cfg.transval_sym_pos):
                transvalent_elements_sum = cfg.transval_sym_pos
            elif (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == cfg.transval_sym_neg):
                transvalent_elements_sum = 0.0
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == cfg.transval_sym_neg):
                transvalent_elements_sum = cfg.transval_sym_neg
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == cfg.transval_sym_pos):
                transvalent_elements_sum = 0.0

            elif (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == 0):
                transvalent_elements_sum = cfg.transval_sym_pos
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == 0):
                transvalent_elements_sum = cfg.transval_sym_neg
            elif (p.expr0[1] == 0) and (p.expr0[1] == 0):
                transvalent_elements_sum = 0
            elif (p.expr0[1] == 0) and (p.expr0[1] == cfg.transval_sym_pos):
                transvalent_elements_sum = cfg.transval_sym_pos
            elif (p.expr0[1] == 0) and (p.expr0[1] == cfg.transval_sym_neg):
                transvalent_elements_sum = cfg.transval_sym_neg

            return (p.expr0[0] + p.expr1[0], transvalent_elements_sum)


    @_('term')
    def expr(self, p):
        """
        Defines the processing of a single term.
        """
        return p.term


    @_('expr TIMES expr')
    def expr(self, p):
        """
        Specifies how the multiplication operation is evaluated
        within the context of transvalent mathematics.
        """

        # ------------------------------------------------------------------
        # Process multiplication of lone elements (e.g., a transvalent 
        # symbol *or* float with another such variable), generating a
        # well-formed transvalent tuple as output.
        # ------------------------------------------------------------------

        # Process "float × float".
        if isinstance(p.expr0, float) and isinstance(p.expr1, float):
            return (p.expr0 * p.expr1, 0.0)

        # Process "float × positive transvalent".
        elif isinstance(p.expr0, float) and (p.expr1 == cfg.transval_sym_pos):
            if (p.expr0 > 0):
                return (0.0, cfg.transval_sym_pos)
            elif (p.expr0 == 0):
                return (1.0, 0.0)
            elif (p.expr0 < 0):
                return (0.0, cfg.transval_sym_neg)

        # Process "float × negative transvalent".
        elif isinstance(p.expr0, float) and (p.expr1 == cfg.transval_sym_neg):
            if (p.expr0 > 0):
                return (0.0, cfg.transval_sym_neg)
            elif (p.expr0 == 0):
                return (-1.0, 0.0)
            elif (p.expr0 < 0):
                return (0.0, cfg.transval_sym_pos)

        # Process "positive transvalent × float".
        elif (p.expr0 == cfg.transval_sym_pos) and isinstance(p.expr1, float):
            if (p.expr1 > 0):
                return (0.0, cfg.transval_sym_pos)
            elif (p.expr1 == 0):
                return (1.0, 0.0)
            elif (p.expr1 < 0):
                return (0.0, cfg.transval_sym_neg)

        # Process "negative transvalent × float".
        elif (p.expr0 == cfg.transval_sym_neg) and isinstance(p.expr1, float):
            if (p.expr1 > 0):
                return (0.0, cfg.transval_sym_neg)
            elif (p.expr1 == 0):
                return (-1.0, 0.0)
            elif (p.expr1 < 0):
                return (0.0, cfg.transval_sym_pos)

        # Process "positive transvalent × positive transvalent".
        elif (p.expr0 == cfg.transval_sym_pos) and (p.expr1 == cfg.transval_sym_pos):
            return (0.0, cfg.transval_sym_pos)

        # Process "positive transvalent × negative transvalent".
        elif (p.expr0 == cfg.transval_sym_pos) and (p.expr1 == cfg.transval_sym_neg):
            return (0.0, cfg.transval_sym_neg)

        # Process "negative transvalent × positive transvalent".
        elif (p.expr0 == cfg.transval_sym_neg) and (p.expr1 == cfg.transval_sym_pos):
            return (0.0, cfg.transval_sym_neg)

        # Process "negative transvalent × negative transvalent".
        elif (p.expr0 == cfg.transval_sym_neg) and (p.expr1 == cfg.transval_sym_neg):
            return (0.0, cfg.transval_sym_pos)


        # ------------------------------------------------------------------
        # Process multiplication of a lone element and a tuple, generating
        # a well-formed transvalent tuple as output.
        # ------------------------------------------------------------------

        # Process "float × tuple".
        elif isinstance(p.expr0, float) and isinstance(p.expr1, tuple):
            return (p.expr0 * p.expr1[0], p.expr1[1])

        # Process "tuple × float".
        elif isinstance(p.expr0, tuple) and isinstance(p.expr1, float):
            return (p.expr0[0] * p.expr1, p.expr0[1])

        # Process "positive transvalent × tuple".
        elif (p.expr0 == cfg.transval_sym_pos) and isinstance(p.expr1, tuple):
            if p.expr1[1] == cfg.transval_sym_pos:
                if p.expr1[0] > 0:
                    return (0.0, cfg.transval_sym_pos)
                elif p.expr1[0] == 0:
                    return (1.0, cfg.transval_sym_pos)
                elif p.expr1[0] < 0:
                    return (0.0, 0.0)
            elif p.expr1[1] == cfg.transval_sym_neg:
                if p.expr1[0] > 0:
                    return (0.0, 0.0)
                elif p.expr1[0] == 0:
                    return (1.0, cfg.transval_sym_neg)
                elif p.expr1[0] < 0:
                    return (0.0, cfg.transval_sym_neg)
            elif p.expr1[1] == 0:
                if p.expr1[0] > 0:
                    return (0.0, cfg.transval_sym_pos)
                elif p.expr1[0] == 0:
                    return (1.0, 0.0)
                elif p.expr1[0] < 0:
                    return (0.0, cfg.transval_sym_neg)

        # Process "negative transvalent × tuple".
        elif (p.expr0 == cfg.transval_sym_neg) and isinstance(p.expr1, tuple):
            if p.expr1[1] == cfg.transval_sym_pos:
                if p.expr1[0] > 0:
                    return (0.0, cfg.transval_sym_neg)
                elif p.expr1[0] == 0:
                    return (-1.0, cfg.transval_sym_neg)
                elif p.expr1[0] < 0:
                    return (0.0, 0.0)
            elif p.expr1[1] == cfg.transval_sym_neg:
                if p.expr1[0] > 0:
                    return (0.0, 0.0)
                elif p.expr1[0] == 0:
                    return (-1.0, cfg.transval_sym_pos)
                elif p.expr1[0] < 0:
                    return (0.0, cfg.transval_sym_pos)
            elif p.expr1[1] == 0:
                if p.expr1[0] > 0:
                    return (0.0, cfg.transval_sym_neg)
                elif p.expr1[0] == 0:
                    return (-1.0, 0.0)
                elif p.expr1[0] < 0:
                    return (0.0, cfg.transval_sym_pos)

        # Process "tuple × positive transvalent".
        elif isinstance(p.expr0, tuple) and (p.expr1 == cfg.transval_sym_pos):
            if p.expr0[1] == cfg.transval_sym_pos:
                if p.expr0[0] > 0:
                    return (0.0, cfg.transval_sym_pos)
                elif p.expr0[0] == 0:
                    return (1.0, cfg.transval_sym_pos)
                elif p.expr0[0] < 0:
                    return (0.0, 0.0)
            elif p.expr0[1] == cfg.transval_sym_neg:
                if p.expr0[0] > 0:
                    return (0.0, 0.0)
                elif p.expr0[0] == 0:
                    return (1.0, cfg.transval_sym_neg)
                elif p.expr0[0] < 0:
                    return (0.0, cfg.transval_sym_neg)
            elif p.expr0[1] == 0:
                if p.expr0[0] > 0:
                    return (0.0, cfg.transval_sym_pos)
                elif p.expr0[0] == 0:
                    return (1.0, 0.0)
                elif p.expr0[0] < 0:
                    return (0.0, cfg.transval_sym_neg)

        # Process "tuple × negative transvalent".
        elif isinstance(p.expr0, tuple) and (p.expr1 == cfg.transval_sym_neg):
            if p.expr0[1] == cfg.transval_sym_pos:
                if p.expr0[0] > 0:
                    return (0.0, cfg.transval_sym_neg)
                elif p.expr0[0] == 0:
                    return (-1.0, cfg.transval_sym_neg)
                elif p.expr0[0] < 0:
                    return (0.0, 0.0)
            elif p.expr0[1] == cfg.transval_sym_neg:
                if p.expr0[0] > 0:
                    return (0.0, 0.0)
                elif p.expr0[0] == 0:
                    return (-1.0, cfg.transval_sym_pos)
                elif p.expr0[0] < 0:
                    return (0.0, cfg.transval_sym_pos)
            elif p.expr0[1] == 0:
                if p.expr0[0] > 0:
                    return (0.0, cfg.transval_sym_neg)
                elif p.expr0[0] == 0:
                    return (-1.0, 0.0)
                elif p.expr0[0] < 0:
                    return (0.0, cfg.transval_sym_pos)

        # ------------------------------------------------------------------
        # Process multiplication of two tuples.
        # ------------------------------------------------------------------

        elif isinstance(p.expr0, tuple) and isinstance(p.expr1, tuple):

            # Where expr0[1] == Ƿ and expr1[1] == Ƿ.
            if (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == cfg.transval_sym_pos):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (
                        (p.expr0[0] > 0) and (p.expr1[0] == 0) \
                        or (p.expr0[0] == 0) and (p.expr1[0] > 0)
                        ):
                    return (1.0, cfg.transval_sym_pos)
                elif (
                        (p.expr0[0] > 0) and (p.expr1[0] < 0) \
                        or (p.expr0[0] < 0) and (p.expr1[0] > 0)
                        ):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (2.0, cfg.transval_sym_pos)
                elif (
                        (p.expr0[0] == 0) and (p.expr1[0] < 0) \
                        or (p.expr0[0] < 0) and (p.expr1[0] == 0) 
                        ):
                    return (1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)

            # Where expr0[1] == Ƿ and expr1[1] == -Ƿ.
            elif (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == cfg.transval_sym_neg):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] > 0) and (p.expr1[0] == 0):
                    return (1.0, cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] > 0):
                    return (-1.0, 0.0)
                elif (p.expr0[0] > 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] < 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (0.0, cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] < 0):
                    return (-1.0, cfg.transval_sym_neg)
                elif (p.expr0[0] < 0) and (p.expr1[0] == 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)

            # Where expr0[1] == -Ƿ and expr1[1] == Ƿ.
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == cfg.transval_sym_pos):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] > 0) and (p.expr1[0] == 0):
                    return (-1.0, cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] > 0):
                    return (1.0, cfg.transval_sym_neg)
                elif (p.expr0[0] > 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] < 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (0.0, cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] < 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] == 0):
                    return (-1.0, cfg.transval_sym_neg)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)

            # Where expr0[1] == -Ƿ and expr1[1] == -Ƿ.
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == cfg.transval_sym_neg):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (
                        (p.expr0[0] > 0) and (p.expr1[0] == 0) \
                        or (p.expr0[0] == 0) and (p.expr1[0] > 0)
                        ):
                    return (-1.0, 0.0)
                elif (
                        (p.expr0[0] > 0) and (p.expr1[0] < 0) \
                        or (p.expr0[0] < 0) and (p.expr1[0] > 0)
                        ):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (-2.0, cfg.transval_sym_pos)
                elif (
                        (p.expr0[0] == 0) and (p.expr1[0] < 0) \
                        or (p.expr0[0] < 0) and (p.expr1[0] == 0) 
                        ):
                    return (-1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)

            # Where expr0[1] == Ƿ and expr1[1] == 0.
            elif (p.expr0[1] == cfg.transval_sym_pos) and (p.expr0[1] == 0):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] > 0) and (p.expr1[0] == 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] == 0) and (p.expr1[0] > 0):
                    return (0.0, cfg.transval_sym_pos)
                elif (p.expr0[0] > 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] < 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] == 0) and (p.expr1[0] < 0):
                    return (0.0, cfg.transval_sym_neg)
                elif (p.expr0[0] < 0) and (p.expr1[0] == 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)

            # Where expr0[1] == 0 and expr1[1] == Ƿ.
            elif (p.expr0[1] == 0) and (p.expr0[1] == cfg.transval_sym_pos):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] > 0) and (p.expr1[0] == 0):
                    return (0.0, cfg.transval_sym_pos)
                elif (p.expr0[0] == 0) and (p.expr1[0] > 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] > 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] < 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] == 0) and (p.expr1[0] < 0):
                    return (1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] == 0):
                    return (0.0, cfg.transval_sym_neg)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)

            # Where expr0[1] == 0 and expr1[1] == -Ƿ.
            elif (p.expr0[1] == 0) and (p.expr0[1] == cfg.transval_sym_neg):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] > 0) and (p.expr1[0] == 0):
                    return (0.0, cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] > 0):
                    return (-1.0, 0.0)
                elif (p.expr0[0] > 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] < 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (-1.0, 0.0)
                elif (p.expr0[0] == 0) and (p.expr1[0] < 0):
                    return (-1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] == 0):
                    return (0.0, cfg.transval_sym_pos)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)

            # Where expr0[1] == -Ƿ and expr1[1] == 0.
            elif (p.expr0[1] == cfg.transval_sym_neg) and (p.expr0[1] == 0):
                if (p.expr0[0] > 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] > 0) and (p.expr1[0] == 0):
                    return (-1.0, 0.0)
                elif (p.expr0[0] == 0) and (p.expr1[0] > 0):
                    return (0.0, cfg.transval_sym_neg)
                elif (p.expr0[0] > 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)
                elif (p.expr0[0] < 0) and (p.expr1[0] > 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_neg)
                elif (p.expr0[0] == 0) and (p.expr1[0] == 0):
                    return (-1.0, 0.0)
                elif (p.expr0[0] == 0) and (p.expr1[0] < 0):
                    return (0.0, cfg.transval_sym_pos)
                elif (p.expr0[0] < 0) and (p.expr1[0] == 0):
                    return (-1.0, 0.0)
                elif (p.expr0[0] < 0) and (p.expr1[0] < 0):
                    return (p.expr0[0] * p.expr1[0], cfg.transval_sym_pos)

            # Where expr0[1] == 0 and expr1[1] == 0.
            elif (p.expr0[1] == 0) and (p.expr0[1] == 0):
                return (p.expr0[0] * p.expr1[0], 0.0)


    """
    @_('term DIVIDE factor')
    def expr(self, p):
        return p.term / p.factor
    """


    @_('factor')
    def term(self, p):
        """
        Defines the processing of a single factor.
        """
        return p.factor


    @_('LPAREN expr RPAREN')
    def factor(self, p):
        """
        Defines the processing of an expression enclosed in parentheses.
        """
        return p.expr


    @_('NUM')
    def factor(self, p):
        """
        Defines the processing of a recognized numerical token.
        """
        return float(p.NUM)


    @_('TRANSVALENT_SYMBOL_INPUT')
    def factor(self, p):
        """
        Defines the processing of a recognized transvalent symbol.
        """
        return cfg.transval_sym_pos


    @_('ID')
    def factor(self, p):
        """
        Checks whether inputted text is an established variable.
        """
        return self.getVar(p.ID)

    def getVar(self, name):
        """
        Assigns the value to a new variable from via user input.
        """
        if name not in self.vars:
            self.vars[name] = float(input(f"Please enter the desired value for {name}: "))
        return self.vars[name]


# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Formatter (to convert transvalent tuples or text for display)
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

def format_result_for_display(
    result_unformatted
    ):
    """
    Formats calculated transvalent tuples of the type "(1.0, Ƿ)" as
    output in the form of "1.0 + Ƿ", for display. Passes conventional
    text strings (e.g., intro or help text) without alteration.
    """

    # For debugging purposes.
    #print("result_unformatted: ", result_unformatted)
    #print("result_unformatted type: ", type(result_unformatted))


    # If the result is a lone real number...
    if isinstance(result_unformatted, float):
        result_formatted = str(result_unformatted) + " + (" + cfg.transval_sym_pos + " - " + cfg.transval_sym_pos + ")"

    # If the result is Ƿ or -Ƿ by itself...
    elif isinstance(result_unformatted, str):
        if result_unformatted == cfg.transval_sym_pos:
            result_formatted = str(0.0) + " + " + cfg.transval_sym_pos
        elif result_unformatted == cfg.transval_sym_neg:
            result_formatted = str(0.0) + " - " + cfg.transval_sym_pos
        else:
            result_formatted = result_unformatted

    # If the result is a well-formed tuple...
    elif isinstance(result_unformatted, tuple):

        # If the transvalent element == Ƿ.
        if result_unformatted[1] == cfg.transval_sym_pos:
            result_formatted = str(float(result_unformatted[0])) + " + " + cfg.transval_sym_pos

        # If the transvalent element == -Ƿ.
        elif result_unformatted[1] == cfg.transval_sym_neg:
            result_formatted = str(float(result_unformatted[0])) + " - " + cfg.transval_sym_pos

        # If the transvalent element == 0.
        elif result_unformatted[1] == 0:
            result_formatted = str(float(result_unformatted[0])) + " + (" + cfg.transval_sym_pos + " - " + cfg.transval_sym_pos + ")"

        # If result_formatted is a string rather than a tuple 
        # (e.g., in the case of the intro text or help text):
        elif isinstance(result_unformatted, str):
            result_formatted = result_unformatted

    return result_formatted


# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Display of pre-prepared text blocks
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

def print_intro_text():
    """
    Displays the standard text block used when launching the package.
    """
    print(cfg.intro_text)


def print_help_text():
    """
    Displays a help file when requested by the user.
    """

    print(cfg.help_text)
    return "('Help' contents displayed above. Returning to command prompt.)"


# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■
# █ Overall logic for executing the module
# ■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■■

def handler(signum, frame):
    """
    Allows the user the option of cleanly exiting the program when 
    input of "Ctrl+C" has been detected.
    """

    # Ask for confirmation that the user wishes to exit the program.
    pre_exit_message = "Quit Liniarote? Please press 'y' to exit this program."
    print(pre_exit_message, end="", flush=True)
    user_response = readchar.readchar()

    # Exit if either "Y" or "y" is inputted as the user's response.
    if user_response.lower() == "y":
        print("\n              Thank you for using Liniarote!\n")
        exit(1)
    else:
        print("", end="\r", flush=True)


if __name__ == '__main__':

    # Display the intro text.
    print_intro_text()

    # Create the lexer.
    lexer = LiniaroteLexer()

    # Check for user input of "Ctrl+C".
    signal.signal(signal.SIGINT, handler)

    while True:

        # Display the command prompt that accepts user input.
        text = input("<LINIAROTE:>  ")

        try:
            # For debugging purposes.
            #tokens_for_display = lexer.tokenize(text)
            #for token in tokens_for_display:
            #    print('token type: ', token.type, "; token value: ", token.value)

            # Process the user's input.
            tokens = lexer.tokenize(text)
            parser = LiniaroteParser()
            result = parser.parse(tokens)

        except:
            result = ("A problem occurred during processing. Please try reformulating your input.")

        # Format the calculated output for display and display it
        # on a special "output" line.
        result = format_result_for_display(result)
        print("    output =  " + result)

        # After calculated output has been displayed, wait
        # before providing a new (blank) command prompt to
        # receive new input from the user.
        time.sleep(0.4)

# ••••-••••-••••-••••-••••-••••-••••--••••-••••-••••-••••-••••-••••-••••


import setuptools

setuptools.setup(
    name="liniarote",
    version="0.2.28",
    url="https://github.com/NeuraXenetica/liniarote",
    author="Matthew E. Gladden",
    author_email="matthew.gladden@neuraxenetica.com",
    description="The Liniarote package is a command-line interface for the Liniarote programming language, for use in performing operations in transvalent mathematics (i.e., in which division by zero is permitted and generates predictable results, displayed with the aid of the transvalent number symbol “Ƿ”).",
    packages=['liniarote'],
    entry_points={'console_scripts': ['liniarote=liniarote.run:main']},
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=[
        'readchar', 'sly'
    ],
)